@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # <Lettuce - Behaviour Driven Development for python>
-# Copyright (C) <2010-2011>  Gabriel Falcão <gabriel@nacaolivre.org>
+# Copyright (C) <2010-2012>  Gabriel Falcão <gabriel@nacaolivre.org>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,7 +17,9 @@
 
 import re
 import codecs
+from fuzzywuzzy import fuzz
 import unicodedata
+from itertools import chain
 from copy import deepcopy
 from lettuce import strings
 from lettuce import languages
@@ -89,6 +91,10 @@ class Language(object):
         name = re.sub(r'^first_of_', u'', attr)
         return unicode(getattr(self, name, u'').split(u"|")[0])
 
+    @property
+    def non_capturable_scenario_separator(self):
+        return re.sub(r'^[(]', '(?:', self.scenario_separator)
+
     @classmethod
     def guess_from_string(cls, string):
         match = re.search(u"language:[ ]*([^\s]+)", string)
@@ -491,14 +497,19 @@ class Scenario(object):
     indentation = 2
     table_indentation = indentation + 2
 
-    def __init__(self, name, remaining_lines, keys, outlines, with_file=None,
-                 original_string=None, language=None):
+    def __init__(self, name, remaining_lines, keys, outlines,
+                 with_file=None,
+                 original_string=None,
+                 language=None,
+                 previous_scenario=None):
 
+        self.feature = None
         if not language:
             language = language()
 
         self.name = name
         self.language = language
+        self.remaining_lines = remaining_lines
         self.steps = self._parse_remaining_lines(remaining_lines,
                                                  with_file,
                                                  original_string)
@@ -507,6 +518,8 @@ class Scenario(object):
         self.with_file = with_file
         self.original_string = original_string
 
+        self.previous_scenario = previous_scenario
+
         if with_file and original_string:
             scenario_definition = ScenarioDescription(self, with_file,
                                                       original_string,
@@ -517,6 +530,11 @@ class Scenario(object):
             self.steps, self.outlines, with_file, original_string))
         self._add_myself_to_steps()
 
+        if original_string and '@' in self.original_string:
+            self.tags = self._find_tags_in(original_string)
+        else:
+            self.tags = None
+
     @property
     def max_length(self):
         if self.outlines:
@@ -561,6 +579,46 @@ class Scenario(object):
     def __repr__(self):
         return u'<Scenario: "%s">' % self.name
 
+    def matches_tags(self, tags):
+        if tags is None:
+            return True
+
+        if not self.tags:
+            return False
+
+        matched = []
+
+        for tag in self.tags:
+            if tag in tags:
+                return True
+
+        for tag in tags:
+            exclude = tag.startswith('-')
+            if exclude:
+                tag = tag[1:]
+
+            fuzzable = tag.startswith('~')
+            if fuzzable:
+                tag = tag[1:]
+
+            result = tag in self.tags
+            if fuzzable:
+                fuzzed = []
+                for internal_tag in self.tags:
+                    ratio = fuzz.ratio(tag, internal_tag)
+                    if exclude:
+                        fuzzed.append(ratio <= 80)
+                    else:
+                        fuzzed.append(ratio > 80)
+
+                result = any(fuzzed)
+            elif exclude:
+                result = tag not in self.tags
+
+            matched.append(result)
+
+        return all(matched)
+
     @property
     def evaluated(self):
         for outline in self.outlines:
@@ -627,6 +685,39 @@ class Scenario(object):
         for step in self.solved_steps:
             step.scenario = self
 
+    def _find_tags_in(self, original_string):
+        broad_regex = re.compile(ur"([@].*)%s: (%s)" % (
+            self.language.scenario_separator,
+            self.name), re.DOTALL)
+
+        regexes = []
+        if not self.previous_scenario:
+            regexes.append(broad_regex)
+
+        else:
+            regexes.append(re.compile(ur"(?:%s: %s.*)([@]?.*)%s: (%s)" % (
+                self.language.non_capturable_scenario_separator,
+                self.previous_scenario.name,
+                self.language.scenario_separator,
+                self.name), re.DOTALL))
+
+        def try_finding_with(regex):
+            found = regex.search(original_string)
+            if found:
+                tag_lines = found.group().splitlines()
+                return list(chain(*map(self._extract_tag, tag_lines)))
+
+        for regex in regexes:
+            found = try_finding_with(regex)
+            if found:
+                return found
+
+        return []
+
+    def _extract_tag(self, item):
+        regex = re.compile(r'[@](\S+)')
+        return regex.findall(item)
+
     def _resolve_steps(self, steps, outlines, with_file, original_string):
         for outline in outlines:
             for step in steps:
@@ -653,17 +744,40 @@ class Scenario(object):
         else:
             prefix = make_prefix(self.language.first_of_scenario)
 
-        head = prefix + self.name
+        head_parts = []
+        if self.tags:
+            tags = ['@%s' % t for t in self.tags]
+            head_parts.append(u' ' * self.indentation)
+            head_parts.append(' '.join(tags) + '\n')
 
-        return strings.rfill(head, self.feature.max_length + 1, append=u'# %s:%d\n' % (self.described_at.file, self.described_at.line))
+        head_parts.append(prefix + self.name)
+
+        head = ''.join(head_parts)
+        appendix = ''
+        if self.described_at:
+            fmt = (self.described_at.file, self.described_at.line)
+            appendix = u'# %s:%d\n' % fmt
+
+        max_length = self.max_length
+        if self.feature:
+            max_length = self.feature.max_length
+
+        return strings.rfill(
+            head, max_length + 1,
+            append=appendix)
 
     def represent_examples(self):
         lines = strings.dicts_to_string(self.outlines, self.keys).splitlines()
         return "\n".join([(u" " * self.table_indentation) + line for line in lines]) + '\n'
 
     @classmethod
-    def from_string(new_scenario, string, with_file=None, original_string=None, language=None):
+    def from_string(new_scenario, string,
+                    with_file=None,
+                    original_string=None,
+                    language=None,
+                    previous_scenario=None):
         """ Creates a new scenario from string"""
+
         # ignoring comments
         string = "\n".join(strings.get_stripped_lines(string, ignore_lines_starting_with='#'))
 
@@ -693,6 +807,7 @@ class Scenario(object):
             with_file=with_file,
             original_string=original_string,
             language=language,
+            previous_scenario=previous_scenario,
         )
 
         return scenario
@@ -780,7 +895,7 @@ class Feature(object):
                 'Features must have a name. e.g: "Feature: This is my name"')
 
         while lines:
-            matched = re.search(r'%s:(.*)' % language.feature, lines[0], re.I)
+            matched = re.search(r'(?:%s):(.*)' % language.feature, lines[0], re.I)
             if matched:
                 name = matched.groups()[0].strip()
                 break
@@ -807,12 +922,18 @@ class Feature(object):
     def _set_definition(self, definition):
         self.described_at = definition
 
+    def _strip_next_scenario_tags(self, string):
+        regex = re.compile(ur'(?:\s*[@]\S+\s*)+$', re.DOTALL)
+        stripped = regex.sub('', string)
+        return stripped or string
+
     def _parse_remaining_lines(self, lines, original_string, with_file=None):
         # replacing occurrences of Scenario Outline, with just "Scenario"
         joined = u"\n".join(lines[1:])
         scenario_prefix = u'%s:' % self.language.first_of_scenario
         regex = re.compile(
-            u"%s:\s" % self.language.scenario_separator, re.U | re.I)
+            u"%s:\s" % self.language.scenario_separator, re.U | re.I | re.DOTALL)
+
         joined = regex.sub(scenario_prefix, joined)
 
         parts = strings.split_wisely(joined, scenario_prefix)
@@ -823,8 +944,9 @@ class Feature(object):
             description = parts[0]
             parts.pop(0)
 
-        scenario_strings = [u"%s: %s" % (self.language.first_of_scenario, s) \
-                            for s in parts if s.strip()]
+        prefix = self.language.first_of_scenario
+        upcoming_scenarios = [
+            u"%s: %s" % (prefix, s) for s in parts if s.strip()]
 
         kw = dict(
             original_string=original_string,
@@ -832,11 +954,27 @@ class Feature(object):
             language=self.language,
         )
 
-        scenarios = [Scenario.from_string(s, **kw) for s in scenario_strings]
+        scenarios = []
+        while upcoming_scenarios:
+            current = self._strip_next_scenario_tags(upcoming_scenarios.pop(0))
+
+            previous_scenario = None
+            has_previous = len(scenarios) > 0
+
+            if has_previous:
+                previous_scenario = scenarios[-1]
+
+            params = dict(
+                previous_scenario=previous_scenario,
+            )
+
+            params.update(kw)
+            current_scenario = Scenario.from_string(current, **params)
+            scenarios.append(current_scenario)
 
         return scenarios, description
 
-    def run(self, scenarios=None, ignore_case=True):
+    def run(self, scenarios=None, ignore_case=True, tags=None):
         call_hook('before_each', 'feature', self)
         scenarios_ran = []
 
@@ -850,6 +988,9 @@ class Feature(object):
             if scenarios_to_run and (index + 1) not in scenarios_to_run:
                 continue
 
+            if not scenario.matches_tags(tags):
+                continue
+
             scenarios_ran.extend(scenario.run(ignore_case))
 
         call_hook('after_each', 'feature', self)
